@@ -6,25 +6,27 @@ a simple interface for sending commands and receiving responses over both MQTT
 and local connections, preferring local when available.
 """
 
+import asyncio
 import logging
 from collections.abc import Callable
 from typing import Any, Protocol, TypeVar, overload
 
 from roborock.containers import RoborockBase
+from roborock.exceptions import RoborockException
 from roborock.protocols.v1_protocol import (
     CommandType,
     ParamsType,
+    RequestMessage,
     SecurityData,
-    create_mqtt_payload_encoder,
     decode_rpc_response,
-    encode_local_payload,
 )
-from roborock.roborock_message import RoborockMessage
+from roborock.roborock_message import RoborockMessage, RoborockMessageProtocol
 
 from .local_channel import LocalChannel
 from .mqtt_channel import MqttChannel
 
 _LOGGER = logging.getLogger(__name__)
+_TIMEOUT = 10.0
 
 
 _T = TypeVar("_T", bound=RoborockBase)
@@ -116,7 +118,7 @@ class PayloadEncodedV1RpcChannel(BaseV1RpcChannel):
         self,
         name: str,
         channel: MqttChannel | LocalChannel,
-        payload_encoder: Callable[[CommandType, ParamsType], RoborockMessage],
+        payload_encoder: Callable[[RequestMessage], RoborockMessage],
     ) -> None:
         """Initialize the channel with a raw channel and an encoder function."""
         self._name = name
@@ -131,18 +133,44 @@ class PayloadEncodedV1RpcChannel(BaseV1RpcChannel):
     ) -> Any:
         """Send a command and return a parsed response RoborockBase type."""
         _LOGGER.debug("Sending command (%s): %s, params=%s", self._name, method, params)
-        message = self._payload_encoder(method, params)
-        response = await self._channel.send_message(message)
-        return decode_rpc_response(response)
+        request_message = RequestMessage(method, params=params)
+        message = self._payload_encoder(request_message)
+
+        future: asyncio.Future[dict[str, Any]] = asyncio.Future()
+
+        def find_response(response_message: RoborockMessage) -> None:
+            try:
+                decoded = decode_rpc_response(response_message)
+            except RoborockException:
+                return
+            if decoded.request_id == request_message.request_id:
+                future.set_result(decoded.data)
+
+        unsub = await self._channel.subscribe(find_response)
+        try:
+            await self._channel.publish(message)
+            return await asyncio.wait_for(future, timeout=_TIMEOUT)
+        except TimeoutError as ex:
+            future.cancel()
+            raise RoborockException(f"Command timed out after {_TIMEOUT}s") from ex
+        finally:
+            unsub()
 
 
 def create_mqtt_rpc_channel(mqtt_channel: MqttChannel, security_data: SecurityData) -> V1RpcChannel:
     """Create a V1 RPC channel using an MQTT channel."""
-    payload_encoder = create_mqtt_payload_encoder(security_data)
-    return PayloadEncodedV1RpcChannel("mqtt", mqtt_channel, payload_encoder)
+    return PayloadEncodedV1RpcChannel(
+        "mqtt",
+        mqtt_channel,
+        lambda x: x.encode_message(RoborockMessageProtocol.RPC_REQUEST, security_data=security_data),
+    )
 
 
 def create_combined_rpc_channel(local_channel: LocalChannel, mqtt_rpc_channel: V1RpcChannel) -> V1RpcChannel:
     """Create a V1 RPC channel that combines local and MQTT channels."""
-    local_rpc_channel = PayloadEncodedV1RpcChannel("local", local_channel, encode_local_payload)
+    local_rpc_channel = PayloadEncodedV1RpcChannel(
+        "local",
+        local_channel,
+        lambda x: x.encode_message(RoborockMessageProtocol.GENERAL_REQUEST),
+    )
     return CombinedV1RpcChannel(local_channel, local_rpc_channel, mqtt_rpc_channel)
